import logging
import os
import sys
import tempfile
from functools import cached_property

import earthkit.data as ekd

from ..model import Model
from .api import RemoteAPI

LOG = logging.getLogger(__name__)


class RemoteModel(Model):
    def __init__(self, **kwargs):
        self.cfg = kwargs
        self.cfg["download_assets"] = False

        self.model = self.cfg["model"]
        self.model_version = self.cfg.get("model_version", "latest")
        self._param = {}
        self.api = RemoteAPI()

        if self.model not in self.api.models():
            LOG.error(f"Model '{self.model}' not available on remote server.")
            LOG.error("Rerun the command with --models --remote to list available remote models.")
            sys.exit(1)

        self.load_parameters()

        super().__init__(**self.cfg)

    def __getattr__(self, name):
        return self.get_parameter(name)

    def run(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            input_file = os.path.join(tmpdirname, "input.grib")
            output_file = os.path.join(tmpdirname, "output.grib")
            self.all_fields.save(input_file)

            self.api.input_file = input_file
            self.api.output_file = output_file

            self.api.run(self.cfg)

            ds = ekd.from_source("file", output_file)
            for field in ds:
                self.write(None, template=field)

    def parse_model_args(self, args):
        return None

    def patch_retrieve_request(self, request):
        if not self.remote_has_patch:
            return

        patched = self.api.patch_retrieve_request(self.cfg, request)
        request.update(patched)

    def load_parameters(self):
        params = self.api.metadata(
            self.model,
            self.model_version,
            [
                "expver",
                "version",
                "grid",
                "area",
                "param_level_ml",
                "param_level_pl",
                "param_sfc",
                "lagged",
                "grib_extra_metadata",
                "retrieve",
                "remote_has_patch",  # custom parameter, checks if remote model need patches
            ],
        )
        self._param.update(params)

    def get_parameter(self, name):
        if (param := self._param.get(name)) is not None:
            return param

        _param = self.api.metadata(self.model, self.model_version, name)
        self._param.update(_param)

        return self._param.get(name)

    @cached_property
    def param_level_ml(self):
        return self.get_parameter("param_level_ml") or ([], [])

    @cached_property
    def param_level_pl(self):
        return self.get_parameter("param_level_pl") or ([], [])

    @cached_property
    def param_sfc(self):
        return self.get_parameter("param_sfc") or []

    @cached_property
    def lagged(self):
        return self.get_parameter("lagged") or False

    @cached_property
    def version(self):
        return self.get_parameter("version") or 1

    @cached_property
    def grib_extra_metadata(self):
        return self.get_parameter("grib_extra_metadata") or {}

    @cached_property
    def retrieve(self):
        return self.get_parameter("retrieve") or {}



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

import earthkit.data as ekd

from .base import RequestBasedInput

LOG = logging.getLogger(__name__)


class CdsInput(RequestBasedInput):
    WHERE = "CDS"

    def pl_load_source(self, **kwargs):
        kwargs["product_type"] = "reanalysis"
        return ekd.from_source("cds", "reanalysis-era5-pressure-levels", kwargs)

    def sfc_load_source(self, **kwargs):
        kwargs["product_type"] = "reanalysis"
        return ekd.from_source("cds", "reanalysis-era5-single-levels", kwargs)

    def ml_load_source(self, **kwargs):
        raise NotImplementedError("CDS does not support model levels")



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import argparse
import logging
import os
import shlex
import sys
import earthkit.data as ekd
from .inputs import available_inputs
from .model import Timer
from .model import available_models
from .model import load_model
from .outputs import available_outputs

ekd.settings.set("cache-policy", "user")

LOG = logging.getLogger(__name__)


def _main(argv):
    parser = argparse.ArgumentParser()

    # See https://github.com/pytorch/pytorch/issues/77764
    os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"

    parser.add_argument(
        "--models",
        action="store_true",
        help="List models and exit",
    )

    parser.add_argument(
        "--debug",
        action="store_true",
        help="Turn on debug",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase verbosity",
    )

    parser.add_argument(
        "--retrieve-requests",
        help=("Print mars requests to stdout." "Use --requests-extra to extend or overide the requests. "),
        action="store_true",
    )

    parser.add_argument(
        "--archive-requests",
        help=("Save mars archive requests to FILE." "Use --requests-extra to extend or overide the requests. "),
        metavar="FILE",
    )

    parser.add_argument(
        "--requests-extra",
        help=("Extends the retrieve or archive requests with a list of key1=value1,key2=value."),
    )

    parser.add_argument(
        "--json",
        action="store_true",
        help=("Dump the requests in JSON format."),
    )

    parser.add_argument(
        "--retrieve-fields-type",
        help="Type of field to retrieve. To use with --retrieve-requests.",
        choices=["constants", "prognostics", "all"],
        default="all",
    )

    parser.add_argument(
        "--retrieve-only-one-date",
        help="Only retrieve the last date/time. To use with --retrieve-requests.",
        action="store_true",
    )

    parser.add_argument(
        "--dump-provenance",
        metavar="FILE",
        help=("Dump information for tracking provenance."),
    )

    parser.add_argument(
        "--input",
        default="mars",
        help="Source to use",
        choices=sorted(available_inputs()),
    )

    parser.add_argument(
        "--file",
        help="Source to use if source=file",
    )

    parser.add_argument(
        "--output",
        default="file",
        help="Where to output the results",
        choices=sorted(available_outputs()),
    )

    parser.add_argument(
        "--date",
        default="-1",
        help="For which analysis date to start the inference (default: yesterday)",
    )

    parser.add_argument(
        "--time",
        type=int,
        default=12,
        help="For which analysis time to start the inference (default: 12)",
    )

    parser.add_argument(
        "--assets",
        default=os.environ.get("AI_MODELS_ASSETS", "."),
        help="Path to directory containing the weights and other assets",
    )

    parser.add_argument(
        "--assets-sub-directory",
        help="Load assets from a subdirectory of --assets based on the name of the model.",
        action=argparse.BooleanOptionalAction,
    )

    parser.parse_args(["--no-assets-sub-directory"])

    parser.add_argument(
        "--assets-list",
        help="List the assets used by the model",
        action="store_true",
    )

    parser.add_argument(
        "--download-assets",
        help="Download assets if they do not exists.",
        action="store_true",
    )

    parser.add_argument(
        "--path",
        help="Path where to write the output of the model",
    )

    parser.add_argument(
        "--fields",
        help="Show the fields needed as input for the model",
        action="store_true",
    )

    parser.add_argument(
        "--expver",
        help="Set the experiment version of the model output. Has higher priority than --metadata.",
    )

    parser.add_argument(
        "--class",
        help="Set the 'class' metadata of the model output. Has higher priority than --metadata.",
        metavar="CLASS",
        dest="class_",
    )

    parser.add_argument(
        "--metadata",
        help="Set additional metadata metadata in the model output",
        metavar="KEY=VALUE",
        action="append",
    )

    parser.add_argument(
        "--num-threads",
        type=int,
        default=1,
        help="Number of threads. Only relevant for some models.",
    )

    parser.add_argument(
        "--lead-time",
        type=int,
        default=240,
        help="Length of forecast in hours.",
    )

    parser.add_argument(
        "--hindcast-reference-year",
        help="For encoding hincast-like outputs",
    )

    parser.add_argument(
        "--hindcast-reference-date",
        help="For encoding hincast-like outputs",
    )

    parser.add_argument(
        "--staging-dates",
        help="For encoding hincast-like outputs",
    )

    parser.add_argument(
        "--only-gpu",
        help="Fail if GPU is not available",
        action="store_true",
    )

    parser.add_argument(
        "--deterministic",
        help="Fail if GPU is not available",
        action="store_true",
    )

    # TODO: deprecate that option
    parser.add_argument(
        "--model-version",
        default="latest",
        help="Model version",
    )

    parser.add_argument(
        "--member",
        nargs=1,
        help="member number (control = 0)"
    )

    parser.add_argument(
        "--version",
        action="store_true",
        help="Print ai-models-gfs version and exit",
    )

    if all(arg not in ("--models", "--version") for arg in argv):
        parser.add_argument(
            "model",
            metavar="MODEL",
            choices=available_models() if "--remote" not in argv else None,
            help="The model to run",
        )

    parser.add_argument(
        "--remote",
        help="Enable remote execution, read url and token from ~/.config/ai-models-gfs/api.yaml",
        action="store_true",
        dest="remote_execution",
        default=(os.environ.get("AI_MODELS_REMOTE", "0") == "1"),
    )

    args, unknownargs = parser.parse_known_args(argv)

    if args.version:
        from ai_models_gfs import __version__

        print(__version__)
        sys.exit(0)

    del args.version

    if args.models:
        if args.remote_execution:
            from .remote import RemoteAPI

            api = RemoteAPI()
            models = api.models()
            if len(models) == 0:
                print(f"No remote models available on {api.url}")
                sys.exit(0)
            print(f"Models available on remote server {api.url}")
        else:
            models = available_models()

        for p in sorted(models):
            print(p)
        sys.exit(0)

    if args.assets_sub_directory:
        args.assets = os.path.join(args.assets, args.model)

    if args.path is None:
        args.path = f"{args.model}.grib"

    if args.file is not None:
        args.input = "file"

    if not args.fields and not args.retrieve_requests:
        logging.basicConfig(
            level="DEBUG" if args.debug else "INFO",
            format="%(asctime)s %(levelname)s %(message)s",
        )

    if args.metadata is None:
        args.metadata = []

    args.metadata = dict(kv.split("=") for kv in args.metadata)

    if args.expver is not None:
        args.metadata["expver"] = args.expver

    if args.class_ is not None:
        args.metadata["class"] = args.class_

    if args.requests_extra:
        if not args.retrieve_requests and not args.archive_requests:
            parser.error("You need to specify --retrieve-requests or --archive-requests")

    run(vars(args), unknownargs)

def run(cfg: dict, model_args: list):
    if cfg["remote_execution"]:
        from .remote import RemoteModel

        model = RemoteModel(**cfg, model_args=model_args)
    else:
        model = load_model(cfg["model"], **cfg, model_args=model_args)

    if cfg["fields"]:
        model.print_fields()
        sys.exit(0)

    # This logic is a bit convoluted, but it is for backwards compatibility.
    if cfg["retrieve_requests"] or (cfg["requests_extra"] and not cfg["archive_requests"]):
        model.print_requests()
        sys.exit(0)

    if cfg["assets_list"]:
        model.print_assets_list()
        sys.exit(0)

    try:
        model.run()
    except FileNotFoundError as e:
        LOG.exception(e)
        LOG.error(
            "It is possible that some files required by %s are missing.",
            cfg["model"],
        )
        LOG.error("Rerun the command as:")
        LOG.error(
            "   %s",
            shlex.join([sys.argv[0], "--download-assets"] + sys.argv[1:]),
        )
        sys.exit(1)
        
    model.finalise()

    if cfg["dump_provenance"]:
        with Timer("Collect provenance information"):
            with open(cfg["dump_provenance"], "w") as f:
                prov = model.provenance()
                import json  # import here so it is not listed in provenance

                json.dump(prov, f, indent=4)


def main():
    with Timer("Total time"):
        _main(sys.argv[1:])


if __name__ == "__main__":
    main()



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import base64
import datetime
import json
import logging
import os
import sys
import time
import urllib.request
from collections import defaultdict
from functools import cached_property

import earthkit.data as ekd
import entrypoints
import numpy as np
from earthkit.data.utils.humanize import seconds
from multiurl import download

from .checkpoint import peek
from .inputs import get_input
from .outputs import get_output
from .stepper import Stepper

LOG = logging.getLogger(__name__)


class Timer:
    def __init__(self, title):
        self.title = title
        self.start = time.time()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        elapsed = time.time() - self.start
        LOG.info("%s: %s.", self.title, seconds(elapsed))


class ArchiveCollector:
    UNIQUE = {"date", "hdate", "time", "referenceDate", "type", "stream", "expver"}

    def __init__(self) -> None:
        self.expect = 0
        self.request = defaultdict(set)

    def add(self, field):
        self.expect += 1
        for k, v in field.items():
            self.request[k].add(str(v))
            if k in self.UNIQUE:
                if len(self.request[k]) > 1:
                    raise ValueError(f"Field {field} has different values for {k}: {self.request[k]}")


class Model:
    lagged = False
    assets_extra_dir = None
    retrieve = {}  # Extra parameters for retrieve
    version = 1  # To be overriden in subclasses
    grib_edition = 2  # Default GRIB edition
    grib_extra_metadata = {}  # Extra metadata for grib files

    param_level_ml = ([], [])  # param, level
    param_level_pl = ([], [])  # param, level
    param_sfc = []  # param

    def __init__(self, input, output, download_assets, **kwargs):
        self.input = get_input(input, self, **kwargs)
        self.output = get_output(output, self, **kwargs)

        for k, v in kwargs.items():
            setattr(self, k, v)

        # We need to call it to initialise the default args
        args = self.parse_model_args(self.model_args)
        if args:
            for k, v in vars(args).items():
                setattr(self, k, v)

        if self.assets_sub_directory:
            if self.assets_extra_dir is not None:
                self.assets += self.assets_extra_dir

        LOG.debug("Asset directory is %s", self.assets)

        try:
            self.date = int(self.date)
        except ValueError:
            pass

        if download_assets:
            self.download_assets(**kwargs)

        self.archiving = defaultdict(ArchiveCollector)
        self.created = time.time()

    @cached_property
    def fields_pl(self):
        return self.input.fields_pl

    @cached_property
    def fields_ml(self):
        return self.input.fields_ml

    @cached_property
    def fields_sfc(self):
        return self.input.fields_sfc

    @cached_property
    def all_fields(self):
        return self.input.all_fields

    def write(self, *args, **kwargs):
        self.collect_archive_requests(
            self.output.write(*args, **kwargs, **self.grib_extra_metadata),
        )

    def collect_archive_requests(self, written):
        if self.archive_requests:
            handle, path = written
            if self.hindcast_reference_year or self.hindcast_reference_date:
                # The clone is necessary because the handle
                # does not return always return recently set keys
                handle = handle.clone()

            self.archiving[path].add(handle.as_namespace("mars"))

    def finalise(self):
        self.output.flush()

        if self.archive_requests:
            with open(self.archive_requests, "w") as f:
                json_requests = []

                for path, archive in self.archiving.items():
                    request = dict(expect=archive.expect)
                    if path is not None:
                        request["source"] = f'"{path}"'
                    request.update(archive.request)
                    request.update(self._requests_extra)

                    if self.json:
                        json_requests.append(request)
                    else:
                        self._print_request("archive", request, file=f)

                if json_requests:

                    def json_default(obj):
                        if isinstance(obj, set):
                            if len(obj) > 1:
                                return sorted(list(obj))
                            else:
                                return obj.pop()
                        raise TypeError

                    print(
                        json.dumps(json_requests, separators=(",", ":"), default=json_default, sort_keys=True),
                        file=f,
                    )

    def download_assets(self, **kwargs):
        for file in self.download_files:
            asset = os.path.realpath(os.path.join(self.assets, file))
            if not os.path.exists(asset):
                os.makedirs(os.path.dirname(asset), exist_ok=True)
                LOG.info("Downloading %s", asset)
                
                download_url = self.download_url.format(file=file)
                temp_download = asset + ".download"
                
                if "aurora" not in self.model:
                    # Try using `multiurl` download first
                    download(download_url, temp_download)
                else:
                    urllib.request.urlretrieve(download_url, temp_download)
                
                os.rename(temp_download, asset)

    @property
    def asset_files(self, **kwargs):
        result = []
        for file in self.download_files:
            result.append(os.path.realpath(os.path.join(self.assets, file)))
        return result

    @cached_property
    def device(self):
        import torch

        device = "cpu"

        if torch.backends.mps.is_available() and torch.backends.mps.is_built():
            device = "mps"

        if torch.cuda.is_available() and torch.backends.cuda.is_built():
            device = "cuda"

        LOG.info(
            "Using device '%s'. The speed of inference depends greatly on the device.",
            device.upper(),
        )

        if self.only_gpu:
            if device == "cpu":
                raise RuntimeError("GPU is not available")

        return device

    def torch_deterministic_mode(self):
        import torch

        LOG.info("Setting deterministic mode for PyTorch")
        os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"

        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True
        torch.use_deterministic_algorithms(True)

    @cached_property
    def providers(self):
        import onnxruntime as ort

        available_providers = ort.get_available_providers()
        providers = []
        for n in ["CUDAExecutionProvider", "CPUExecutionProvider"]:
            if n in available_providers:
                providers.append(n)

        LOG.info(
            "Using device '%s'. The speed of inference depends greatly on the device.",
            ort.get_device(),
        )

        if self.only_gpu:
            assert isinstance(ort.get_device(), str)
            if ort.get_device() == "CPU":
                raise RuntimeError("GPU is not available")

            providers = ["CUDAExecutionProvider"]

        LOG.info("ONNXRuntime providers: %s", providers)

        return providers

    def timer(self, title):
        return Timer(title)

    def stepper(self, step):
        # We assume that we call this method only once
        # just before the first iteration.
        elapsed = time.time() - self.created
        LOG.info("Model initialisation: %s", seconds(elapsed))
        return Stepper(step, self.lead_time)

    def _datetimes(self, dates):
        date = self.date
        assert isinstance(date, int)
        if date <= 0:
            date = datetime.datetime.utcnow() + datetime.timedelta(days=date)
            date = date.year * 10000 + date.month * 100 + date.day

        time = self.time
        assert isinstance(time, int)
        if time < 100:
            time *= 100
        # assert time in (0, 600, 1200, 1800), time

        lagged = self.lagged
        if not lagged:
            lagged = [0]

        result = []
        for basedate in dates:
            for lag in lagged:
                date = basedate + datetime.timedelta(hours=lag)
                result.append(
                    (
                        date.year * 10000 + date.month * 100 + date.day,
                        date.hour,
                    ),
                )

        return result

    def datetimes(self, step=0):
        if self.staging_dates:
            assert step == 0, step
            dates = []
            with open(self.staging_dates) as f:
                for line in f:
                    dates.append(datetime.datetime.fromisoformat(line.strip()))

            return self._datetimes(dates)

        date = self.date
        assert isinstance(date, int)
        if date <= 0:
            date = datetime.datetime.utcnow() + datetime.timedelta(days=date)
            date = date.year * 10000 + date.month * 100 + date.day

        time = self.time
        assert isinstance(time, int)
        if time < 100:
            time *= 100

        # assert time in (0, 600, 1200, 1800), time

        full = datetime.datetime(
            date // 10000,
            date % 10000 // 100,
            date % 100,
            time // 100,
            time % 100,
        ) + datetime.timedelta(hours=step)
        return self._datetimes([full])

    def print_fields(self):
        param, level = self.param_level_pl
        print("Grid:", self.grid)
        print("Area:", self.area)
        print("Pressure levels:")
        print("   Levels:", level)
        print("   Params:", param)
        print("Single levels:")
        print("   Params:", self.param_sfc)

    def print_assets_list(self):
        for file in self.download_files:
            print(file)

    def _print_request(self, verb, request, file=sys.stdout):
        r = [verb]
        for k, v in sorted(request.items()):
            if not isinstance(v, (list, tuple, set)):
                v = [v]

            if k in ("area", "grid", "frame", "rotation", "bitmap"):
                v = [str(_) for _ in v]
            else:
                v = [str(_) for _ in sorted(v)]

            v = "/".join(v)
            r.append(f"{k}={v}")

        r = ",\n   ".join(r)
        print(r, file=file)
        print(file=file)

    @property
    def _requests_extra(self):
        if not self.requests_extra:
            return {}
        extra = [_.split("=") for _ in self.requests_extra.split(",")]
        extra = {a: b for a, b in extra}
        return extra

    def print_requests(self):
        requests = self._requests()

        if self.json:
            print(json.dumps(requests, indent=4))
            return

        for r in requests:
            self._print_request("retrieve", r)

    def _requests_unfiltered(self):
        result = []

        first = dict(
            target="input.grib",
            grid=self.grid,
            area=self.area,
        )
        first.update(self.retrieve)

        for date, time in self.datetimes():  # noqa F402
            param, level = self.param_level_pl

            r = dict(
                date=date,
                time=time,
            )
            r.update(first)
            first = {}

            if param and level:

                # PL
                r.update(
                    dict(
                        levtype="pl",
                        levelist=level,
                        param=param,
                        date=date,
                        time=time,
                    )
                )

                r.update(self._requests_extra)

                self.patch_retrieve_request(r)

                result.append(dict(**r))

            # ML
            param, level = self.param_level_ml

            if param and level:
                r.update(
                    dict(
                        levtype="ml",
                        levelist=level,
                        param=param,
                        date=date,
                        time=time,
                    )
                )

                r.update(self._requests_extra)

                self.patch_retrieve_request(r)

                result.append(dict(**r))

            param = self.param_sfc
            if param:
                # SFC
                r.update(
                    dict(
                        levtype="sfc",
                        param=self.param_sfc,
                        date=date,
                        time=time,
                        levelist="off",
                    )
                )

                self.patch_retrieve_request(r)
                result.append(dict(**r))

        return result

    def _requests(self):

        def filter_constant(request):
            # We check for 'sfc' because param 'z' can be ambiguous
            if request.get("levtype") == "sfc":
                param = set(self.constant_fields) & set(request.get("param", []))
                if param:
                    request["param"] = list(param)
                    return True

            return False

        def filter_prognostic(request):
            # TODO: We assume here that prognostic fields are
            # the ones that are not constant. This may not always be true
            if request.get("levtype") == "sfc":
                param = set(request.get("param", [])) - set(self.constant_fields)
                if param:
                    request["param"] = list(param)
                    return True
                return False

            return True

        def filter_last_date(request):
            date, time = max(self.datetimes())
            return request["date"] == date and request["time"] == time

        def noop(request):
            return request

        filter_type = {
            "constants": filter_constant,
            "prognostics": filter_prognostic,
            "all": noop,
        }[self.retrieve_fields_type]

        filter_dates = {
            True: filter_last_date,
            False: noop,
        }[self.retrieve_only_one_date]

        result = []
        for r in self._requests_unfiltered():
            if filter_type(r) and filter_dates(r):
                result.append(r)

        return result

    def patch_retrieve_request(self, request):
        # Overriden in subclasses if needed
        pass

    def peek_into_checkpoint(self, path):
        return peek(path)

    def parse_model_args(self, args):
        if args:
            raise NotImplementedError(f"This model does not accept arguments {args}")

    def provenance(self):
        from .provenance import gather_provenance_info

        return gather_provenance_info(self.asset_files)

    def forcing_and_constants(self, date, param):
        source = self.all_fields[:1]

        ds = ekd.from_source(
            "forcings",
            source,
            date=date,
            param=param,
        )

        assert len(ds) == len(param), (len(ds), len(param), date)

        return ds.to_numpy(dtype=np.float32)

    @cached_property
    def gridpoints(self):
        return len(self.all_fields[0].grid_points()[0])

    @cached_property
    def start_datetime(self):
        for field in self.all_fields.order_by(valid_datetime="ascending"):
            if field['param']!='tp':
                start_datetime = field.datetime()["valid_time"]
        return start_datetime

    @property
    def constant_fields(self):
        raise NotImplementedError("constant_fields")

    def write_input_fields(
        self,
        fields,
        accumulations=None,
        accumulations_template=None,
        accumulations_shape=None,
        ignore=None,
    ):
        LOG.info("Starting date is %s", self.start_datetime)
        LOG.info("Writing input fields")
        if ignore is None:
            ignore = []

        with self.timer("Writing step 0"):
            for field in fields:
                if field.metadata("shortName") in ignore:
                    continue

                if field.datetime()["valid_time"] == self.start_datetime:
                    self.write(
                        None,
                        template=field,
                        step=0,
                    )

            if accumulations is not None:
                if accumulations_template is None:
                    accumulations_template = fields.sel(param="msl")[0]

                if accumulations_shape is None:
                    accumulations_shape = accumulations_template.shape

                if accumulations_template.metadata("edition") == 1:
                    for param in accumulations:

                        self.write(
                            np.zeros(accumulations_shape, dtype=np.float32),
                            stepType="accum",
                            template=accumulations_template,
                            param=param,
                            startStep=0,
                            endStep=0,
                            date=int(self.start_datetime.strftime("%Y%m%d")),
                            time=int(self.start_datetime.strftime("%H%M")),
                            check=True,
                        )
                else:
                    # # TODO: Remove this when accumulations are supported for GRIB edition 2

                    template = """
                    R1JJQv//AAIAAAAAAAAA3AAAABUBAGIAABsBAQfoCRYGAAAAAQAAABECAAEAAQAJBAIwMDAxAAAA
                    SAMAAA/XoAAAAAAG////////////////////AAAFoAAAAtEAAAAA/////wVdSoAAAAAAMIVdSoAV
                    cVlwAAPQkAAD0JAAAAAAOgQAAAAIAcEC//8AAAABAAAAAAH//////////////wfoCRYGAAABAAAA
                    AAECAQAAAAD/AAAAAAAAABUFAA/XoAAAAAAAAIAKAAAAAAAAAAYG/wAAAAUHNzc3N0dSSUL//wAC
                    AAAAAAAAANwAAAAVAQBiAAAbAQEH6AkWDAAAAAEAAAARAgABAAEACQQBMDAwMQAAAEgDAAAP16AA
                    AAAABv///////////////////wAABaAAAALRAAAAAP////8FXUqAAAAAADCFXUqAFXFZcAAD0JAA
                    A9CQAAAAADoEAAAACAHBAv//AAAAAQAAAAAB//////////////8H6AkWDAAAAQAAAAABAgEAAAAA
                    /wAAAAAAAAAVBQAP16AAAAAAAACACgAAAAAAAAAGBv8AAAAFBzc3Nzc=
                    """

                    template = base64.b64decode(template)
                    accumulations_template = ekd.from_source("memory", template)[0]

                    for param in accumulations:
                        self.write(
                            np.zeros(accumulations_shape, dtype=np.float32),
                            stepType="accum",
                            template=accumulations_template,
                            param=param,
                            startStep=0,
                            endStep=0,
                            date=int(self.start_datetime.strftime("%Y%m%d")),
                            time=int(self.start_datetime.strftime("%H%M")),
                            check=True,
                        )


def load_model(name, **kwargs):
    return available_models()[name].load()(**kwargs)


def available_models():
    result = {}
    for e in entrypoints.get_group_all("ai_models_gfs.model"):
        result[e.name] = e
    return result



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property

import earthkit.data as ekd
import earthkit.regrid as ekr
import entrypoints
from earthkit.data.indexing.fieldlist import FieldArray

LOG = logging.getLogger(__name__)


def get_input(name, *args, **kwargs):
    return available_inputs()[name].load()(*args, **kwargs)


def available_inputs():
    result = {}
    for e in entrypoints.get_group_all("ai_models_gfs.input"):
        result[e.name] = e
    return result



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import itertools
import logging
import os

import earthkit.data as ekd
from earthkit.data.core.temporary import temp_file
from earthkit.data.indexing.fieldlist import FieldArray
from multiurl import download

from .base import RequestBasedInput
from .compute import make_z_from_gh
from .interpolate import Interpolate
from .recenter import recenter
from .transform import NewMetadataField

LOG = logging.getLogger(__name__)

CONSTANTS = (
    "z",
    "sdor",
    "slor",
)

CONSTANTS_URL = "https://get.ecmwf.int/repository/test-data/ai-models/opendata/constants-{resol}.grib2"

RESOLS = {
    (0.25, 0.25): ("0p25", (0.25, 0.25), False, False, {}),
    (0.1, 0.1): (
        "0p25",
        (0.25, 0.25),
        True,
        True,
        dict(
            longitudeOfLastGridPointInDegrees=359.9,
            iDirectionIncrementInDegrees=0.1,
            jDirectionIncrementInDegrees=0.1,
            Ni=3600,
            Nj=1801,
        ),
    ),
    # "N320": ("0p25", (0.25, 0.25), True, False, dict(gridType='reduced_gg')),
    # "O96": ("0p25", (0.25, 0.25), True, False, dict(gridType='reduced_gg', )),
}


def _identity(x):
    return x


class OpenDataInput(RequestBasedInput):
    WHERE = "OPENDATA"

    def __init__(self, owner, **kwargs):
        self.owner = owner

    def _adjust(self, kwargs):

        kwargs.setdefault("step", 0)

        if "level" in kwargs:
            # OpenData uses levelist instead of level
            kwargs["levelist"] = kwargs.pop("level")

        if "area" in kwargs:
            kwargs.pop("area")

        grid = kwargs.pop("grid")
        if isinstance(grid, list):
            grid = tuple(grid)

        kwargs["resol"], source, interp, oversampling, metadata = RESOLS[grid]
        r = dict(**kwargs)
        r.update(self.owner.retrieve)

        if interp:

            logging.info("Interpolating input data from %s to %s.", source, grid)
            if oversampling:
                logging.warning("This will oversample the input data.")
            return Interpolate(grid, source, metadata)
        else:
            return _identity

    def pl_load_source(self, **kwargs):

        gh_to_z = _identity
        interpolate = self._adjust(kwargs)

        kwargs["levtype"] = "pl"
        request = kwargs.copy()

        param = [p.lower() for p in kwargs["param"]]
        assert isinstance(param, (list, tuple))

        if "z" in param:
            logging.warning("Parameter 'z' on pressure levels is not available in ECMWF open data, using 'gh' instead")
            param = list(param)
            param.remove("z")
            if "gh" not in param:
                param.append("gh")
            kwargs["param"] = param
            gh_to_z = make_z_from_gh

        logging.info("ecmwf-open-data %s", kwargs)

        opendata = recenter(ekd.from_source("ecmwf-open-data", **kwargs))
        opendata = gh_to_z(opendata)
        opendata = interpolate(opendata)

        return self.check_pl(opendata, request)

    def constants(self, constant_params, request, kwargs):
        if len(constant_params) == 1:
            logging.warning(
                f"Single level parameter '{constant_params[0]}' is"
                " not available in ECMWF open data, using constants.grib2 instead"
            )
        else:
            logging.warning(
                f"Single level parameters {constant_params} are"
                " not available in ECMWF open data, using constants.grib2 instead"
            )

        cachedir = os.path.expanduser("~/.cache/ai-models-gfs")
        constants_url = CONSTANTS_URL.format(resol=request["resol"])
        basename = os.path.basename(constants_url)

        if not os.path.exists(cachedir):
            os.makedirs(cachedir)

        path = os.path.join(cachedir, basename)

        if not os.path.exists(path):
            logging.info("Downloading %s to %s", constants_url, path)
            download(constants_url, path + ".tmp")
            os.rename(path + ".tmp", path)

        ds = ekd.from_source("file", path)
        ds = ds.sel(param=constant_params)

        tmp = temp_file()

        out = ekd.new_grib_output(tmp.path)

        for f in ds:
            out.write(
                f.to_numpy(),
                template=f,
                date=kwargs["date"],
                time=kwargs["time"],
                step=kwargs.get("step", 0),
            )

        out.close()

        result = ekd.from_source("file", tmp.path)
        result._tmp = tmp

        return result

    def sfc_load_source(self, **kwargs):
        interpolate = self._adjust(kwargs)

        kwargs["levtype"] = "sfc"
        request = kwargs.copy()

        param = [p.lower() for p in kwargs["param"]]
        assert isinstance(param, (list, tuple))

        constant_params = []
        param = list(param)
        for c in CONSTANTS:
            if c in param:
                param.remove(c)
                constant_params.append(c)

        if constant_params:
            constants = self.constants(constant_params, request, kwargs)
        else:
            constants = ekd.from_source("empty")

        kwargs["param"] = param

        opendata = recenter(ekd.from_source("ecmwf-open-data", **kwargs))
        opendata = opendata + constants
        opendata = interpolate(opendata)

        # Fix grib2/eccodes bug

        opendata = FieldArray([NewMetadataField(f, levelist=None) for f in opendata])

        return self.check_sfc(opendata, request)

    def ml_load_source(self, **kwargs):
        interpolate = self._adjust(kwargs)
        kwargs["levtype"] = "ml"
        request = kwargs.copy()

        opendata = recenter(ekd.from_source("ecmwf-open-data", **kwargs))
        opendata = interpolate(opendata)

        return self.check_ml(opendata, request)

    def check_pl(self, ds, request):
        self._check(ds, "PL", request, "param", "levelist")
        return ds

    def check_sfc(self, ds, request):
        self._check(ds, "SFC", request, "param")
        return ds

    def check_ml(self, ds, request):
        self._check(ds, "ML", request, "param", "levelist")
        return ds

    def _check(self, ds, what, request, *keys):

        def _(p):
            if len(p) == 1:
                return p[0]

        expected = set()
        for p in itertools.product(*[request[key] for key in keys]):
            expected.add(p)

        found = set()
        for f in ds:
            found.add(tuple(f.metadata(key) for key in keys))

        missing = expected - found
        if missing:
            missing = [_(p) for p in missing]
            if len(missing) == 1:
                raise ValueError(f"The following {what} parameter '{missing[0]}' is not available in ECMWF open data")
            raise ValueError(f"The following {what} parameters {missing} are not available in ECMWF open data")

        extra = found - expected
        if extra:
            extra = [_(p) for p in extra]
            if len(extra) == 1:
                raise ValueError(f"Unexpected {what} parameter '{extra[0]}' from ECMWF open data")
            raise ValueError(f"Unexpected {what} parameters {extra} from ECMWF open data")




# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import os
import pickle
import zipfile
from typing import Any

LOG = logging.getLogger(__name__)


class FakeStorage:
    def __init__(self):
        import torch

        self.dtype = torch.float32
        self._untyped_storage = torch.UntypedStorage(0)


class UnpicklerWrapper(pickle.Unpickler):
    def __init__(self, file, **kwargs):
        super().__init__(file, **kwargs)

    def persistent_load(self, pid: Any) -> Any:
        return FakeStorage()


def tidy(x):
    if isinstance(x, dict):
        return {k: tidy(v) for k, v in x.items()}

    if isinstance(x, list):
        return [tidy(v) for v in x]

    if isinstance(x, tuple):
        return tuple([tidy(v) for v in x])

    if x is None:
        return None

    if isinstance(x, (int, float, str, bool)):
        return x

    return x


def peek(path):
    with zipfile.ZipFile(path, "r") as f:
        data_pkl = None
        for b in f.namelist():
            if os.path.basename(b) == "data.pkl":
                if data_pkl is not None:
                    raise Exception(f"Found two data.pkl files in {path}: {data_pkl} and {b}")
                data_pkl = b

    LOG.info(f"Found data.pkl at {data_pkl}")

    with zipfile.ZipFile(path, "r") as f:
        unpickler = UnpicklerWrapper(f.open(data_pkl, "r"))
        x = tidy(unpickler.load())
        return tidy(x)



from netCDF4 import Dataset as DS
import numpy as np
import pygrib as pg
import datetime

def grib2nc(model,path,lead,step,date,time,inputmodel):

    # Check if the file is still open
    os.system(f"lsof {args.path}")
    print("BYE")
    if "." in path:
        out = path.rsplit(".",1)[0] + ".nc"
    else:
        out = path + ".nc"

    varmap = {
            "u10": ['10 metre U wind component','m s-1']
            ,"v10": ['10 metre V wind component','m s-1']
            ,"t2": ['2 metre temperature','K']
            ,"msl": ['Pressure reduced to MSL','Pa']
            ,"u100": ['100 metre U wind component','m s-1']
            ,"v100": ['100 metre U wind component','m s-1']
            ,"sp": ['Surface pressure','Pa']
            ,"tcwv": ['Precipitable water','kg m-2']
            ,"apcp": ['6-hr accumulated precipitation','m']
            ,"u": ['U component of wind','m s-1']
            ,"v": ['V component of wind','m s-1']
            ,"t": ['Temperature','K']
            ,"z": ['Geopotential','m2 s-2']
            ,"r": ['Relative humidity', '%']
            ,"q": ['Specific humidity','kg kg-1']
            ,"w": ['Vertical velocity','Pa s-1']
            }

    ec2gfsmap = {
            "u":"u"
            ,"v":"v"
            ,"w":"w"
            ,"z":"z"
            ,"q":"q"
            ,"r":"r"
            ,"t":"t"
            ,"10u":"u10"
            ,"10v":"v10"
            ,"100u":"u100"
            ,"100v":"v100"
            ,"2t":"t2"
            ,"msl":"msl"
            ,"sp":"sp"
            ,"tcwv":"tcwv"
            ,"tp":"apcp"
            }

    unique_pl_vars = []
    unique_sfc_vars = []
    levels = []
    ncdata = {}
    print("BYE2")
    print(path)
    grib = pg.open(path)
    print("BYEtest")
    y_shape,x_shape = grib[1].values.shape
    lats,lons = grib[1].latlons()
    lats = lats[:,0]
    lons = lons[0,:]
    print("BYE3")
    grib.seek(0)
    for grb in grib:
        if grb.levelType=="pl":
            if grb.level not in levels:
                levels.append(grb.level)
            if grb.shortName not in unique_pl_vars:
                unique_pl_vars.append(grb.shortName)
        else:
            if grb.shortName not in unique_sfc_vars:
                unique_sfc_vars.append(grb.shortName)
    print("BYE4")
    levels.sort(reverse=True)
    levelmap = {}

    for c,level in enumerate(levels):
        levelmap[level] = c
    grib.seek(0)
    for pl_var in unique_pl_vars:
        modelvar = ec2gfsmap[pl_var]
        ncdata[modelvar] = {
            'values': np.zeros((lead // step + 1, len(levels), y_shape, x_shape)).astype('float32'),
            'name': varmap[modelvar][0], 'units': varmap[modelvar][1]
        }
    for sfc_var in unique_sfc_vars:
        modelvar = ec2gfsmap[sfc_var]
        ncdata[modelvar] = {
            'values': np.zeros((lead // step + 1, y_shape, x_shape)).astype('float32'),
            'name': varmap[modelvar][0], 'units': varmap[modelvar][1]
        }

    grib.seek(0)
    for grb in grib:
        shortName = grb.shortName
        timestep = int(grb.step/step)
        level = grb.level
        levelType = grb.levelType
        if (shortName=='z' and levelType=='sfc') or shortName not in ec2gfsmap.keys():
            continue
        gfsequivalent = ec2gfsmap[shortName]
        vals = grb.values
        if levelType=='pl':
            levelind = levelmap[level]
            ncdata[gfsequivalent]['values'][timestep, levelind, :, :] = vals
        elif levelType=='sfc':
            ncdata[gfsequivalent]['values'][timestep, :, :] = vals

    f = DS(out, 'w', format='NETCDF4')
    f.createDimension('time', lead//step + 1)
    f.createDimension('level', len(levels))
    f.createDimension('longitude', x_shape)
    f.createDimension('latitude', y_shape)

    initdt = datetime.datetime.strptime(f"{date}{time}","%Y%m%d%H00")
    inityr = str(initdt.year)
    initmnth = str(initdt.month).zfill(2)
    initday = str(initdt.day).zfill(2)
    inithr = str(initdt.hour).zfill(2)
    times = []

    times = []
    for i in np.arange(0,lead + step,step):
        times.append(int((initdt + datetime.timedelta(hours=int(i))).timestamp()))
        # Create time, longitude, latitude, and level variables in the NetCDF file
    create_variable_nochunk(
        f, 'time', ('time',), np.array(times), {
            'long_name': 'Date and Time', 'units': 'seconds since 1970-1-1',
            'calendar': 'standard'
        }
    )
    create_variable_nochunk(
        f, 'longitude', ('longitude',), lons, {
            'long_name': 'Longitude', 'units': 'degree'
        }
    )
    create_variable_nochunk(
        f, 'latitude', ('latitude',), lats, {
            'long_name': 'Latitude', 'units': 'degree'
        }
    )
    create_variable_nochunk(
        f, 'level', ('level',), np.array(
            levels
        ), {'long_name': 'Isobaric surfaces', 'units': 'hPa'}
    )

    # Create variables for each meteorological parameter
    for variable in unique_pl_vars + unique_sfc_vars:
        dims = ('time', 'level', 'latitude', 'longitude') if variable in unique_pl_vars else ('time', 'latitude', 'longitude')
        if 'level' in dims:
            chunksizes = (1,1,y_shape,x_shape)
        else:
            chunksizes = (1,y_shape,x_shape)
        create_variable(
            f, variable, dims, ncdata[ec2gfsmap[variable]]['values'], {
                'long_name': ncdata[ec2gfsmap[variable]]['name'], 'units': ncdata[ec2gfsmap[variable]]['units']
            },
            chunksizes
        )

    f.Conventions = 'CF-1.8'
    f.version = '3_2025-02-20'
    f.model_name = model
    f.initialization_model = inputmodel
    f.initialization_time = '%s-%s-%sT%s:00:00' % (inityr,initmnth,initday,inithr)
    f.first_forecast_hour = "0"
    f.last_forecast_hour = f"{lead}"
    f.forecast_hour_step = f"{step}"
    f.creation_time = (datetime.datetime.utcnow()).strftime('%Y-%m-%dT%H:%M:%S')
    f.close()

def create_variable(f, name, dimensions, data, attrs, chunksizes):
    if name in ['time','level']:
        dtype = 'i4'
    else:
        dtype = 'f4'
    var = f.createVariable(name, dtype, dimensions,compression='zlib',complevel=4,chunksizes=chunksizes)
    var[:] = data
    for attr_name, attr_value in attrs.items():
        var.setncattr(attr_name, attr_value)

def create_variable_nochunk(f, name, dimensions, data, attrs):
    if name in ['time','level']:
        dtype = 'i4'
    else:
        dtype = 'f4'
    var = f.createVariable(name, dtype, dimensions,compression='zlib',complevel=4)
    var[:] = data
    for attr_name, attr_value in attrs.items():
        var.setncattr(attr_name, attr_value)




# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import itertools
import logging
import warnings
from functools import cached_property

import earthkit.data as ekd
import entrypoints
import numpy as np

LOG = logging.getLogger(__name__)


class Output:
    def write(self, *args, **kwargs):
        pass

    def flush(self, *args, **kwargs):
        pass


class GribOutputBase(Output):
    def __init__(self, owner, path, metadata, **kwargs):
        self._first = True
        metadata.setdefault("stream", "oper")
        metadata.setdefault("expver", owner.expver)
        metadata.setdefault("class", "ml")

        self.path = path
        self.owner = owner
        self.metadata = metadata

    @cached_property
    def grib_keys(self):
        edition = self.metadata.pop("edition", self.owner.grib_edition)

        _grib_keys = dict(
            edition=edition,
            generatingProcessIdentifier=self.owner.version,
        )
        _grib_keys.update(self.metadata)

        return _grib_keys

    @cached_property
    def output(self):
        return ekd.new_grib_output(
            self.path,
            split_output=True,
            **self.grib_keys,
        )

    def write(self, data, *args, check=False, **kwargs):

        try:
            handle, path = self.output.write(data, *args, **kwargs)

        except Exception:
            if data is not None:
                if np.isnan(data).any():
                    raise ValueError(f"NaN values found in field. args={args} kwargs={kwargs}")
                if np.isinf(data).any():
                    raise ValueError(f"Infinite values found in field. args={args} kwargs={kwargs}")

            options = {}
            options.update(self.grib_keys)
            options.update(kwargs)
            LOG.error("Failed to write data to %s %s", args, options)
            cmd = []
            for k, v in options.items():
                if isinstance(v, (int, str, float)):
                    cmd.append("%s=%s" % (k, v))

            LOG.error("grib_set -s%s", ",".join(cmd))

            raise

        if check:
            # Check that the GRIB keys are as expected

            if kwargs.get("expver") is None:
                ignore = ("template", "check_nans", "expver", "class", "type", "stream")
            else:
                ignore = ("template", "check_nans")

            for key, value in itertools.chain(self.grib_keys.items(), kwargs.items()):
                if key in ignore:
                    continue

                # If "param" is a string, we what to compare it to the shortName
                if key == "param":
                    try:
                        float(value)
                    except ValueError:
                        key = "shortName"

                assert str(handle.get(key)) == str(value), (key, handle.get(key), value)

        return handle, path

class FileOutput(GribOutputBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        LOG.info("Writing results to %s", self.path)


class NoneOutput(Output):
    def __init__(self, *args, **kwargs):
        LOG.info("Results will not be written.")

    def write(self, *args, **kwargs):
        pass


class HindcastReLabel:
    def __init__(self, owner, output, hindcast_reference_year=None, hindcast_reference_date=None, **kwargs):
        self.owner = owner
        self.output = output
        self.hindcast_reference_year = int(hindcast_reference_year) if hindcast_reference_year else None
        self.hindcast_reference_date = int(hindcast_reference_date) if hindcast_reference_date else None
        assert self.hindcast_reference_year is not None or self.hindcast_reference_date is not None

    def write(self, *args, **kwargs):
        if "hdate" in kwargs:
            warnings.warn(f"Ignoring hdate='{kwargs['hdate']}' in HindcastReLabel", stacklevel=3)
            kwargs.pop("hdate")

        if "date" in kwargs:
            warnings.warn(f"Ignoring date='{kwargs['date']}' in HindcastReLabel", stacklevel=3)
            kwargs.pop("date")

        date = kwargs["template"]["date"]
        hdate = kwargs["template"]["hdate"]

        if hdate is not None:
            # Input was a hindcast
            referenceDate = (
                self.hindcast_reference_date
                if self.hindcast_reference_date is not None
                else self.hindcast_reference_year * 10000 + date % 10000
            )
            assert date == referenceDate, (
                date,
                referenceDate,
                hdate,
                kwargs["template"],
            )
            kwargs["referenceDate"] = referenceDate
            kwargs["hdate"] = hdate
        else:
            referenceDate = (
                self.hindcast_reference_date
                if self.hindcast_reference_date is not None
                else self.hindcast_reference_year * 10000 + date % 10000
            )
            kwargs["referenceDate"] = referenceDate
            kwargs["hdate"] = date

        kwargs.setdefault("check", True)

        return self.output.write(*args, **kwargs)

    def flush(self, *args, **kwargs):
        return self.output.flush(*args, **kwargs)


class NoLabelling:

    def __init__(self, owner, output, **kwargs):
        self.owner = owner
        self.output = output

    def write(self, *args, **kwargs):
        kwargs["deleteLocalDefinition"] = 1
        return self.output.write(*args, **kwargs)

    def flush(self, *args, **kwargs):
        return self.output.flush(*args, **kwargs)


def get_output(name, owner, *args, **kwargs):
    result = available_outputs()[name].load()(owner, *args, **kwargs)
    if kwargs.get("hindcast_reference_year") is not None or kwargs.get("hindcast_reference_date") is not None:
        result = HindcastReLabel(owner, result, **kwargs)
    if owner.expver is None:
        result = NoLabelling(owner, result, **kwargs)
    return result


def available_outputs():
    result = {}
    for e in entrypoints.get_group_all("ai_models_gfs.output"):
        result[e.name] = e
    return result



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import itertools
import logging
import os
import eccodes
import numpy as np
import earthkit.data as ekd
import urllib.request
import tempfile
from datetime import datetime,timedelta
from scipy.interpolate import RegularGridInterpolator
from .base import RequestBasedInput

LOG = logging.getLogger(__name__)

class GfsInput(RequestBasedInput):
    WHERE = "GFS"
    def pl_load_source(self, **kwargs):
        interp = bool(int(kwargs['grid'][0]))
        # Use a named temp file to ensure it persists
        temp_path = os.path.join(tempfile.gettempdir(), "sample_pres.grib")
        sample_url = "https://noaa-oar-mlwp-data.s3.amazonaws.com/sample_files/sample_pres.grib"

        # Check if the file already exists before downloading
        if not os.path.exists(temp_path):
            print(f"Downloading {sample_url} to {temp_path}...")
            urllib.request.urlretrieve(sample_url, temp_path)
        else:
            print(f"File already exists: {temp_path}, skipping download.")

        # Download the file to the temp directory
        urllib.request.urlretrieve(sample_url, temp_path)

        # Load the GRIB file from the temp directory
        sample_pressure_grib = ekd.from_source("file", temp_path)

        # Create a new GRIB output file for the formatted pressure data
        formatted_pressure_file = (
            f"/tmp/ai-models-gfs/gfspresformatted_{str(kwargs['date'])}_"
            f"{str(kwargs['time']).zfill(2)}.grib"
        )
        if not os.path.isdir("/tmp/ai-models-gfs"):
            os.makedirs("/tmp/ai-models-gfs")

        formatted_pressure_output = ekd.new_grib_output(
            formatted_pressure_file, edition=1
        )
        # Construct the URL to fetch GFS pressure data
        gfs_pressure_url = (
            f"https://noaa-gfs-bdp-pds.s3.amazonaws.com/gfs.{str(kwargs['date'])}/"
            f"{str(kwargs['time']).zfill(2)}/atmos/"
            f"gfs.t{str(kwargs['time']).zfill(2)}z.pgrb2.0p25.f000"
        )
        # Load the GFS pressure data from the URL
        gfs_pressure_data = ekd.from_source("url", gfs_pressure_url)

        # Iterate over the sample pressure GRIB messages
        for grib_message in sample_pressure_grib:
            parameter_name = grib_message['shortName']
            pressure_level = grib_message['level']
            template = grib_message
            # Set the date and time for the GRIB template
            eccodes.codes_set(template.handle._handle, "date", int(kwargs['date']))
            eccodes.codes_set(
                template.handle._handle, "time", int(kwargs['time']) * 100
            )

            if parameter_name == "z":
                # Select geopotential height data and convert to meters
                geopotential_height_data = gfs_pressure_data.sel(
                    param="gh", level=pressure_level
                )
                data_array = geopotential_height_data[0].to_numpy() * 9.80665
            else:
                # Select other parameters' data
                parameter_data = gfs_pressure_data.sel(
                    param=parameter_name, level=pressure_level
                )
                data_array = parameter_data[0].to_numpy()
            if interp:
                data_array = interpolate(
                                        data_array,
                                        np.arange(90,-90.25,-0.25),
                                        np.arange(0,360,0.25),
                                        np.arange(90,-91,-1),
                                        np.arange(0,360,1)
                                        )
            if interp:
                template = set_eccodes(template)
            # Write the data to the formatted GRIB file using the template
            formatted_pressure_output.write(data_array, template=template)

        # Load the formatted GRIB file and return it
        formatted_pressure_grib = ekd.from_source("file", formatted_pressure_file)
        return formatted_pressure_grib

    def sfc_load_source(self, **kwargs):
        interp = bool(int(kwargs['grid'][0]))

        # Use a named temp file to ensure it persists
        temp_path = os.path.join(tempfile.gettempdir(), "sample_sfc.grib")
        sample_url = "https://noaa-oar-mlwp-data.s3.amazonaws.com/sample_files/sample_sfc.grib"

        # Check if the file already exists before downloading
        if not os.path.exists(temp_path):
            print(f"Downloading {sample_url} to {temp_path}...")
            urllib.request.urlretrieve(sample_url, temp_path)
        else:
            print(f"File already exists: {temp_path}, skipping download.")

        # Download the file to the temp directory
        urllib.request.urlretrieve(sample_url, temp_path)

        # Load the GRIB file from the temp directory
        sample_surface_grib = ekd.from_source("file", temp_path)

        # Create a new GRIB output file for the formatted surface data
        formatted_surface_file = (
            f"/tmp/ai-models-gfs/gfssfcformatted_{str(kwargs['date'])}_"
            f"{str(kwargs['time']).zfill(2)}.grib"
        )

        if not os.path.isdir("/tmp/ai-models-gfs"):
            os.makedirs("/tmp/ai-models-gfs")

        formatted_surface_output = ekd.new_grib_output(
            formatted_surface_file, edition=1
        )
        # Construct the URL to fetch GFS surface data
        gfs_surface_url = (
            f"https://noaa-gfs-bdp-pds.s3.amazonaws.com/gfs.{str(kwargs['date'])}/"
            f"{str(kwargs['time']).zfill(2)}/atmos/"
            f"gfs.t{str(kwargs['time']).zfill(2)}z.pgrb2.0p25.f000"
        )
        # Load the GFS surface data from the URL
        gfs_surface_data = ekd.from_source("url", gfs_surface_url)

        # Iterate over the sample surface GRIB messages
        for grib_message in sample_surface_grib:
            parameter_name = grib_message['shortName']
            surface_level = grib_message['level']
            template = grib_message

            # Set the date and time for the GRIB template
            eccodes.codes_set(template.handle._handle, "date", int(kwargs['date']))
            eccodes.codes_set(
                template.handle._handle, "time", int(kwargs['time']) * 100
            )
            if parameter_name == "tp":
                # For total precipitation, create an array of zeros
                data_array = np.zeros((721, 1440))
            elif parameter_name in ["z", "lsm"]:
                # For geopotential height and land-sea mask, use the data directly
                data_array = grib_message.to_numpy()
            elif parameter_name == "msl":
                # Select mean sea level pressure data
                mean_sea_level_pressure_data = gfs_surface_data.sel(param="prmsl")
                data_array = mean_sea_level_pressure_data[0].to_numpy()
            elif parameter_name == "tcwv":
                # Select total column water vapor data
                total_column_water_vapor_data = gfs_surface_data.sel(param="pwat")
                data_array = total_column_water_vapor_data[0].to_numpy()
            else:
                # Select other parameters' data
                parameter_data = gfs_surface_data.sel(param=parameter_name)
                data_array = parameter_data[0].to_numpy()
            if interp:
                data_array = interpolate(
                                        data_array,
                                        np.arange(90,-90.25,-0.25),
                                        np.arange(0,360,0.25),
                                        np.arange(90,-91,-1),
                                        np.arange(0,360,1)
                                        )
            if interp:
                template = set_eccodes(template)            
            # Write the data to the formatted GRIB file using the template
            formatted_surface_output.write(data_array, template=template)

        # Load the formatted GRIB file and return it
        formatted_surface_grib = ekd.from_source("file", formatted_surface_file)
        return formatted_surface_grib

    def ml_load_source(self, **kwargs):
        raise NotImplementedError("CDS does not support model levels")

class GefsInput(RequestBasedInput):
    WHERE = "GEFS"

    def __init__(self, owner, **kwargs):
        super().__init__(owner, **kwargs)
        self.kwargs = kwargs

    def pl_load_source(self, **kwargs):
        interp = bool(int(kwargs['grid'][0]))
        member = str(self.kwargs['member'][0]).zfill(2)
        if member=='00':
            member = 'c00'
        else:
            member = f'p{member}'

        # Use a named temp file to ensure it persists
        temp_path = os.path.join(tempfile.gettempdir(), "sample_pres.grib")
        sample_url = "https://noaa-oar-mlwp-data.s3.amazonaws.com/sample_files/sample_pres.grib"

        # Check if the file already exists before downloading
        if not os.path.exists(temp_path):
            print(f"Downloading {sample_url} to {temp_path}...")
            urllib.request.urlretrieve(sample_url, temp_path)
        else:
            print(f"File already exists: {temp_path}, skipping download.")

        # Download the file to the temp directory
        urllib.request.urlretrieve(sample_url, temp_path)

        # Load the GRIB file from the temp directory
        sample_pressure_grib = ekd.from_source("file", temp_path)

        # Create a new GRIB output file for the formatted pressure data
        formatted_pressure_file = (
            f"/tmp/ai-models-gfs/gefspresformatted_{str(kwargs['date'])}_"
            f"{str(kwargs['time']).zfill(2)}_{member}.grib"
        )

        if not os.path.isdir("/tmp/ai-models-gfs"):
            os.makedirs("/tmp/ai-models-gfs")

        formatted_pressure_output = ekd.new_grib_output(
            formatted_pressure_file, edition=1
        )
        # Construct the URL to fetch GFS pressure data

        gefs_pressure_url_a = (
            f"https://noaa-gefs-pds.s3.amazonaws.com/gefs.{str(kwargs['date'])}/"
            f"{str(kwargs['time']).zfill(2)}/atmos/pgrb2ap5/"
            f"ge{member}.t{str(kwargs['time']).zfill(2)}z.pgrb2a.0p50.f000"
        )

        gefs_pressure_url_b = (
            f"https://noaa-gefs-pds.s3.amazonaws.com/gefs.{str(kwargs['date'])}/"
            f"{str(kwargs['time']).zfill(2)}/atmos/pgrb2bp5/"
            f"ge{member}.t{str(kwargs['time']).zfill(2)}z.pgrb2b.0p50.f000"
        )
        print(gefs_pressure_url_a)
        # Load the GFS pressure data from the URL
        gefs_pressure_data_a = ekd.from_source("url", gefs_pressure_url_a)
        gefs_pressure_data_b = ekd.from_source("url", gefs_pressure_url_b)
        gefs_pressure_data = gefs_pressure_data_a + gefs_pressure_data_b

        # Iterate over the sample pressure GRIB messages
        for grib_message in sample_pressure_grib:
            parameter_name = grib_message['shortName']
            pressure_level = grib_message['level']
            template = grib_message
            # Set the date and time for the GRIB template
            eccodes.codes_set(template.handle._handle, "date", int(kwargs['date']))
            eccodes.codes_set(
                template.handle._handle, "time", int(kwargs['time']) * 100
            )

            if parameter_name == "z":
                # Select geopotential height data and convert to meters
                geopotential_height_data = gefs_pressure_data.sel(
                    param="gh", level=pressure_level
                )
                data_array = geopotential_height_data[0].to_numpy() * 9.80665
            else:
                # Select other parameters' data
                parameter_data = gefs_pressure_data.sel(
                    param=parameter_name, level=pressure_level
                )
                data_array = parameter_data[0].to_numpy()

            if interp:
                data_array = interpolate(
                                        data_array,
                                        np.arange(90,-90.50,-0.50),
                                        np.arange(0,360,0.50),
                                        np.arange(90,-91,-1),
                                        np.arange(0,360,1)
                                        )
            else:
                data_array = interpolate(
                                        data_array,
                                        np.arange(90,-90.50,-0.50),
                                        np.arange(0,360,0.50),
                                        np.arange(90,-90.25,-0.25),
                                        np.arange(0,360,0.25)
                                        )
            # Write the data to the formatted GRIB file using the template
            if interp:
                template = set_eccodes(template)
            formatted_pressure_output.write(data_array, template=template)

        # Load the formatted GRIB file and return it
        formatted_pressure_grib = ekd.from_source("file", formatted_pressure_file)
        return formatted_pressure_grib

    def sfc_load_source(self, **kwargs):
        interp = bool(int(kwargs['grid'][0]))
        member = str(self.kwargs['member'][0]).zfill(2)
        if member=='00':
            member = 'c00'
        else:
            member = f'p{member}'

        # Use a named temp file to ensure it persists
        temp_path = os.path.join(tempfile.gettempdir(), "sample_sfc.grib")
        sample_url = "https://noaa-oar-mlwp-data.s3.amazonaws.com/sample_files/sample_sfc.grib"

        # Check if the file already exists before downloading
        if not os.path.exists(temp_path):
            print(f"Downloading {sample_url} to {temp_path}...")
            urllib.request.urlretrieve(sample_url, temp_path)
        else:
            print(f"File already exists: {temp_path}, skipping download.")

        # Download the file to the temp directory
        urllib.request.urlretrieve(sample_url, temp_path)

        # Load the GRIB file from the temp directory
        sample_surface_grib = ekd.from_source("file", temp_path)

        # Create a new GRIB output file for the formatted surface data
        formatted_surface_file = (
            f"/tmp/ai-models-gfs/gfssfcformatted_{str(kwargs['date'])}_"
            f"{str(kwargs['time']).zfill(2)}_{member}.grib"
        )

        if not os.path.isdir("/tmp/ai-models-gfs"):
            os.makedirs("/tmp/ai-models-gfs")

        formatted_surface_output = ekd.new_grib_output(
            formatted_surface_file, edition=1
        )

        # Construct the URL to fetch GFS surface data
        gefs_surface_url_a = (
            f"https://noaa-gefs-pds.s3.amazonaws.com/gefs.{str(kwargs['date'])}/"
            f"{str(kwargs['time']).zfill(2)}/atmos/pgrb2ap5/"
            f"ge{member}.t{str(kwargs['time']).zfill(2)}z.pgrb2a.0p50.f000"
        )

        # Construct the URL to fetch GFS surface data
        gefs_surface_url_b = (
            f"https://noaa-gefs-pds.s3.amazonaws.com/gefs.{str(kwargs['date'])}/"
            f"{str(kwargs['time']).zfill(2)}/atmos/pgrb2bp5/"
            f"ge{member}.t{str(kwargs['time']).zfill(2)}z.pgrb2b.0p50.f000"
        )

        # Load the GFS surface data from the URL
        gefs_surface_data_a = ekd.from_source("url", gefs_surface_url_a)
        gefs_surface_data_b = ekd.from_source("url", gefs_surface_url_b)
        gefs_surface_data = gefs_surface_data_a + gefs_surface_data_b
        # Iterate over the sample surface GRIB messages

        for grib_message in sample_surface_grib:
            parameter_name = grib_message['shortName']
            surface_level = grib_message['level']
            template = grib_message

            # Set the date and time for the GRIB template
            eccodes.codes_set(template.handle._handle, "date", int(kwargs['date']))
            eccodes.codes_set(
                template.handle._handle, "time", int(kwargs['time']) * 100
            )

            if parameter_name == "tp":
                # For total precipitation, create an array of zeros
                data_array = np.zeros((721, 1440))
                if interp:
                    data_array = interpolate(
                                            data_array,
                                            np.arange(90,-90.25,-0.25),
                                            np.arange(0,360,0.25),
                                            np.arange(90,-91,-1),
                                            np.arange(0,360,1)
                                            )
            elif parameter_name in ["z", "lsm"]:
                # For geopotential height and land-sea mask, use the data directly
                data_array = grib_message.to_numpy()
                if interp:
                    data_array = interpolate(
                                            data_array,
                                            np.arange(90,-90.25,-0.25),
                                            np.arange(0,360,0.25),
                                            np.arange(90,-91,-1),
                                            np.arange(0,360,1)
                                            )
            elif parameter_name == "msl":
                # Select mean sea level pressure data
                mean_sea_level_pressure_data = gefs_surface_data.sel(param="prmsl")
                data_array = mean_sea_level_pressure_data[0].to_numpy()
                if interp:
                    data_array = interpolate(
                                            data_array,
                                            np.arange(90,-90.50,-0.50),
                                            np.arange(0,360,0.50),
                                            np.arange(90,-91,-1),
                                            np.arange(0,360,1)
                                            )
                else:
                    data_array = interpolate(
                                            data_array,
                                            np.arange(90,-90.50,-0.50),
                                            np.arange(0,360,0.50),
                                            np.arange(90,-90.25,-0.25),
                                            np.arange(0,360,0.25)
                                            )
            elif parameter_name == "tcwv":
                # Select total column water vapor data
                total_column_water_vapor_data = gefs_surface_data.sel(param="pwat")
                data_array = total_column_water_vapor_data[0].to_numpy()
                if interp:
                    data_array = interpolate(
                                            data_array,
                                            np.arange(90,-90.50,-0.50),
                                            np.arange(0,360,0.50),
                                            np.arange(90,-91,-1),
                                            np.arange(0,360,1)
                                            )
                else:
                    data_array = interpolate(
                                            data_array,
                                            np.arange(90,-90.50,-0.50),
                                            np.arange(0,360,0.50),
                                            np.arange(90,-90.25,-0.25),
                                            np.arange(0,360,0.25)
                                            )
            else:
                # Select other parameters' data
                parameter_data = gefs_surface_data.sel(param=parameter_name)
                data_array = parameter_data[0].to_numpy()
                if interp:
                    data_array = interpolate(
                                            data_array,
                                            np.arange(90,-90.50,-0.50),
                                            np.arange(0,360,0.50),
                                            np.arange(90,-91,-1),
                                            np.arange(0,360,1)
                                            )
                else:
                    data_array = interpolate(
                                            data_array,
                                            np.arange(90,-90.50,-0.50),
                                            np.arange(0,360,0.50),
                                            np.arange(90,-90.25,-0.25),
                                            np.arange(0,360,0.25)
                                            )



            # Write the data to the formatted GRIB file using the template
            if interp:
                template = set_eccodes(template)
            formatted_surface_output.write(data_array, template=template)

        # Load the formatted GRIB file and return it
        formatted_surface_grib = ekd.from_source("file", formatted_surface_file)
        return formatted_surface_grib

    def ml_load_source(self, **kwargs):
        raise NotImplementedError("CDS does not support model levels")

class GdasInput(RequestBasedInput):
    WHERE = "GDAS"

    def pl_load_source(self, **kwargs):
        interp = bool(int(kwargs['grid'][0]))
        # Load the sample pressure GRIB file
        # Use a named temp file to ensure it persists
        temp_path = os.path.join(tempfile.gettempdir(), "sample_pres.grib")
        sample_url = "https://noaa-oar-mlwp-data.s3.amazonaws.com/sample_files/sample_pres.grib"

        # Check if the file already exists before downloading
        if not os.path.exists(temp_path):
            print(f"Downloading {sample_url} to {temp_path}...")
            urllib.request.urlretrieve(sample_url, temp_path)
        else:
            print(f"File already exists: {temp_path}, skipping download.")

        # Download the file to the temp directory
        urllib.request.urlretrieve(sample_url, temp_path)

        # Load the GRIB file from the temp directory
        sample_pressure_grib = ekd.from_source("file", temp_path)

        # Create a new GRIB output file for the formatted pressure data
        formatted_pressure_file = (
            f"/tmp/ai-models-gfs/gdaspresformatted_{str(kwargs['date'])}_"
            f"{str(kwargs['time']).zfill(2)}.grib"
        )
        if not os.path.isdir("/tmp/ai-models-gfs"):
            os.makedirs("/tmp/ai-models-gfs")
        formatted_pressure_output = ekd.new_grib_output(
            formatted_pressure_file, edition=1
        )
        # Construct the URL to fetch GFS pressure data
        gdas_pressure_url = (
            f"https://noaa-gfs-bdp-pds.s3.amazonaws.com/gdas.{str(kwargs['date'])}/"
            f"{str(kwargs['time']).zfill(2)}/atmos/"
            f"gdas.t{str(kwargs['time']).zfill(2)}z.pgrb2.0p25.f000"
        )
        print(gdas_pressure_url)
        # Load the GFS pressure data from the URL
        gdas_pressure_data = ekd.from_source("url", gdas_pressure_url)

        # Iterate over the sample pressure GRIB messages
        for grib_message in sample_pressure_grib:
            parameter_name = grib_message['shortName']
            pressure_level = grib_message['level']
            template = grib_message

            # Set the date and time for the GRIB template
            eccodes.codes_set(template.handle._handle, "date", int(kwargs['date']))
            eccodes.codes_set(
                template.handle._handle, "time", int(kwargs['time']) * 100
            )

            if parameter_name == "z":
                # Select geopotential height data and convert to meters
                geopotential_height_data = gdas_pressure_data.sel(
                    param="gh", level=pressure_level
                )
                data_array = geopotential_height_data[0].to_numpy() * 9.80665
            else:
                # Select other parameters' data
                parameter_data = gdas_pressure_data.sel(
                    param=parameter_name, level=pressure_level
                )
                data_array = parameter_data[0].to_numpy()

            if interp:
                data_array = interpolate(
                                        data_array,
                                        np.arange(90,-90.25,-0.25),
                                        np.arange(0,360,0.25),
                                        np.arange(90,-91,-1),
                                        np.arange(0,360,1)
                                        )
            # Write the data to the formatted GRIB file using the template
            if interp:
                template = set_eccodes(template)
            formatted_pressure_output.write(data_array, template=template)

        # Load the formatted GRIB file and return it
        formatted_pressure_grib = ekd.from_source("file", formatted_pressure_file)
        return formatted_pressure_grib

    def sfc_load_source(self, **kwargs):
        interp = bool(int(kwargs['grid'][0]))
        # Use a named temp file to ensure it persists
        temp_path = os.path.join(tempfile.gettempdir(), "sample_sfc.grib")
        sample_url = "https://noaa-oar-mlwp-data.s3.amazonaws.com/sample_files/sample_sfc.grib"

        # Check if the file already exists before downloading
        if not os.path.exists(temp_path):
            print(f"Downloading {sample_url} to {temp_path}...")
            urllib.request.urlretrieve(sample_url, temp_path)
        else:
            print(f"File already exists: {temp_path}, skipping download.")

        # Download the file to the temp directory
        urllib.request.urlretrieve(sample_url, temp_path)

        # Load the GRIB file from the temp directory
        sample_surface_grib = ekd.from_source("file", temp_path)

        # Create a new GRIB output file for the formatted surface data
        formatted_surface_file = (
            f"/tmp/ai-models-gfs/gdassfcformatted_{str(kwargs['date'])}_"
            f"{str(kwargs['time']).zfill(2)}.grib"
        )
        if not os.path.isdir("/tmp/ai-models-gfs"):
            os.makedirs("/tmp/ai-models-gfs")
        formatted_surface_output = ekd.new_grib_output(
            formatted_surface_file, edition=1
        )
        # Construct the URL to fetch GFS surface data
        gdas_surface_url = (
            f"https://noaa-gfs-bdp-pds.s3.amazonaws.com/gdas.{str(kwargs['date'])}/"
            f"{str(kwargs['time']).zfill(2)}/atmos/"
            f"gdas.t{str(kwargs['time']).zfill(2)}z.pgrb2.0p25.f000"
        )
        # Load the GFS surface data from the URL
        gdas_surface_data = ekd.from_source("url", gdas_surface_url)

        # Iterate over the sample surface GRIB messages
        for grib_message in sample_surface_grib:
            parameter_name = grib_message['shortName']
            surface_level = grib_message['level']
            template = grib_message

            # Set the date and time for the GRIB template
            eccodes.codes_set(template.handle._handle, "date", int(kwargs['date']))
            eccodes.codes_set(
                template.handle._handle, "time", int(kwargs['time']) * 100
            )

            if parameter_name == "tp":
                # For total precipitation, create an array of zeros
                data_array = np.zeros((721, 1440))
            elif parameter_name in ["z", "lsm"]:
                # For geopotential height and land-sea mask, use the data directly
                data_array = grib_message.to_numpy()
            elif parameter_name == "msl":
                # Select mean sea level pressure data
                mean_sea_level_pressure_data = gdas_surface_data.sel(param="prmsl")
                data_array = mean_sea_level_pressure_data[0].to_numpy()
            elif parameter_name == "tcwv":
                # Select total column water vapor data
                total_column_water_vapor_data = gdas_surface_data.sel(param="pwat")
                data_array = total_column_water_vapor_data[0].to_numpy()
            else:
                # Select other parameters' data
                parameter_data = gdas_surface_data.sel(param=parameter_name)
                data_array = parameter_data[0].to_numpy()

            if interp:
                data_array = interpolate(
                                        data_array,
                                        np.arange(90,-90.25,-0.25),
                                        np.arange(0,360,0.25),
                                        np.arange(90,-91,-1),
                                        np.arange(0,360,1)
                                        )
            # Write the data to the formatted GRIB file using the template
            if interp:
                template = set_eccodes(template)
            formatted_surface_output.write(data_array, template=template)

        # Load the formatted GRIB file and return it
        formatted_surface_grib = ekd.from_source("file", formatted_surface_file)
        return formatted_surface_grib

    def ml_load_source(self, **kwargs):
        raise NotImplementedError("CDS does not support model levels")

def interpolate(data,inlats,inlons,outlats,outlons):
    inlons_extended = np.concatenate(([inlons[-1] - 360], inlons, [inlons[0] + 360]))
    data_extended = np.concatenate((data[:, -1:], data, data[:, :1]), axis=1)
    interpolator = RegularGridInterpolator((inlats, inlons_extended), data_extended, bounds_error=False)
    outlon_grid,outlat_grid = np.meshgrid(outlons,outlats)
    points = np.array([outlat_grid.flatten(),outlon_grid.flatten()]).T
    data_interpolated = interpolator(points).reshape(outlat_grid.shape)
    return data_interpolated

def set_eccodes(template):
    grib_handle = template.handle._handle
    eccodes.codes_set(grib_handle, "Ni", 360)  # Longitude points
    eccodes.codes_set(grib_handle, "Nj", 181)  # Latitude points

    # Set correct grid spacing for 1-degree resolution
    eccodes.codes_set(grib_handle, "iDirectionIncrementInDegrees", 1.0)
    eccodes.codes_set(grib_handle, "jDirectionIncrementInDegrees", 1.0)

    # Define latitude/longitude bounds
    eccodes.codes_set(grib_handle, "latitudeOfFirstGridPointInDegrees", 90)
    eccodes.codes_set(grib_handle, "longitudeOfFirstGridPointInDegrees", 0)
    eccodes.codes_set(grib_handle, "latitudeOfLastGridPointInDegrees", -90)
    eccodes.codes_set(grib_handle, "longitudeOfLastGridPointInDegrees", 359)
    return template



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

import earthkit.data as ekd

from .base import RequestBasedInput

LOG = logging.getLogger(__name__)


class MarsInput(RequestBasedInput):
    WHERE = "MARS"

    def __init__(self, owner, **kwargs):
        self.owner = owner

    def pl_load_source(self, **kwargs):
        kwargs["levtype"] = "pl"
        logging.debug("load source mars %s", kwargs)
        return ekd.from_source("mars", kwargs)

    def sfc_load_source(self, **kwargs):
        kwargs["levtype"] = "sfc"
        logging.debug("load source mars %s", kwargs)
        return ekd.from_source("mars", kwargs)

    def ml_load_source(self, **kwargs):
        kwargs["levtype"] = "ml"
        logging.debug("load source mars %s", kwargs)
        return ekd.from_source("mars", kwargs)



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

import earthkit.data as ekd
import numpy as np
import tqdm
from earthkit.data.core.temporary import temp_file

LOG = logging.getLogger(__name__)

CHECKED = set()


def _init_recenter(ds, f):

    # For now, we only support the 0.25x0.25 grid from OPENDATA (centered on the greenwich meridian)

    latitudeOfFirstGridPointInDegrees = f.metadata("latitudeOfFirstGridPointInDegrees")
    longitudeOfFirstGridPointInDegrees = f.metadata("longitudeOfFirstGridPointInDegrees")
    latitudeOfLastGridPointInDegrees = f.metadata("latitudeOfLastGridPointInDegrees")
    longitudeOfLastGridPointInDegrees = f.metadata("longitudeOfLastGridPointInDegrees")
    iDirectionIncrementInDegrees = f.metadata("iDirectionIncrementInDegrees")
    jDirectionIncrementInDegrees = f.metadata("jDirectionIncrementInDegrees")
    scanningMode = f.metadata("scanningMode")
    Ni = f.metadata("Ni")
    Nj = f.metadata("Nj")

    assert scanningMode == 0
    assert latitudeOfFirstGridPointInDegrees == 90
    assert longitudeOfFirstGridPointInDegrees == 180
    assert latitudeOfLastGridPointInDegrees == -90
    assert longitudeOfLastGridPointInDegrees == 179.75
    assert iDirectionIncrementInDegrees == 0.25
    assert jDirectionIncrementInDegrees == 0.25

    assert Ni == 1440
    assert Nj == 721

    shape = (Nj, Ni)
    roll = -Ni // 2
    axis = 1

    key = (
        latitudeOfFirstGridPointInDegrees,
        longitudeOfFirstGridPointInDegrees,
        latitudeOfLastGridPointInDegrees,
        longitudeOfLastGridPointInDegrees,
        iDirectionIncrementInDegrees,
        jDirectionIncrementInDegrees,
        Ni,
        Nj,
    )

    ############################

    if key not in CHECKED:
        lon = ekd.from_source("forcings", ds, param=["longitude"], date=f.metadata("date"))[0]
        assert np.all(np.roll(lon.to_numpy(), roll, axis=axis)[:, 0] == 0)
        CHECKED.add(key)

    return (shape, roll, axis, dict(longitudeOfFirstGridPointInDegrees=0, longitudeOfLastGridPointInDegrees=359.75))


def recenter(ds):

    tmp = temp_file()

    out = ekd.new_grib_output(tmp.path)

    for f in tqdm.tqdm(ds, delay=0.5, desc="Recentering", leave=False):

        shape, roll, axis, metadata = _init_recenter(ds, f)

        data = f.to_numpy()
        assert data.shape == shape, (data.shape, shape)

        data = np.roll(data, roll, axis=axis)

        out.write(data, template=f, **metadata)

    out.close()

    result = ekd.from_source("file", tmp.path)
    result._tmp = tmp

    return result



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

import earthkit.data as ekd
import tqdm
from earthkit.data.core.temporary import temp_file
from earthkit.data.indexing.fieldlist import FieldArray

LOG = logging.getLogger(__name__)

G = 9.80665  # Same a pgen


def make_z_from_gh(ds):

    tmp = temp_file()

    out = ekd.new_grib_output(tmp.path)
    other = []

    for f in tqdm.tqdm(ds, delay=0.5, desc="GH to Z", leave=False):

        if f.metadata("param") == "gh":
            out.write(f.to_numpy() * G, template=f, param="z")
        else:
            other.append(f)

    out.close()

    result = FieldArray(other) + ekd.from_source("file", tmp.path)
    result._tmp = tmp

    return result



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

import earthkit.data as ekd
import earthkit.regrid as ekr
import tqdm
from earthkit.data.core.temporary import temp_file

LOG = logging.getLogger(__name__)


class Interpolate:
    def __init__(self, grid, source, metadata):
        self.grid = list(grid) if isinstance(grid, tuple) else grid
        self.source = list(source) if isinstance(source, tuple) else source
        self.metadata = metadata

    def __call__(self, ds):
        tmp = temp_file()

        out = ekd.new_grib_output(tmp.path)

        result = []
        for f in tqdm.tqdm(ds, delay=0.5, desc="Interpolating", leave=False):
            data = ekr.interpolate(f.to_numpy(), dict(grid=self.source), dict(grid=self.grid))
            out.write(data, template=f, **self.metadata)

        out.close()

        result = ekd.from_source("file", tmp.path)
        result._tmp = tmp

        print("Interpolated data", tmp.path)

        return result



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property

import earthkit.data as ekd

LOG = logging.getLogger(__name__)


class RequestBasedInput:
    def __init__(self, owner, **kwargs):
        self.owner = owner

    def _patch(self, **kargs):
        r = dict(**kargs)
        self.owner.patch_retrieve_request(r)
        return r

    @cached_property
    def fields_sfc(self):
        param = self.owner.param_sfc
        if not param:
            return ekd.from_source("empty")

        LOG.info(f"Loading surface fields from {self.WHERE}")

        return ekd.from_source(
            "multi",
            [
                self.sfc_load_source(
                    **self._patch(
                        date=date,
                        time=time,
                        param=param,
                        grid=self.owner.grid,
                        area=self.owner.area,
                        **self.owner.retrieve,
                    )
                )
                for date, time in self.owner.datetimes()
            ],
        )

    @cached_property
    def fields_pl(self):
        param, level = self.owner.param_level_pl
        if not (param and level):
            return ekd.from_source("empty")

        LOG.info(f"Loading pressure fields from {self.WHERE}")
        return ekd.from_source(
            "multi",
            [
                self.pl_load_source(
                    **self._patch(
                        date=date,
                        time=time,
                        param=param,
                        level=level,
                        grid=self.owner.grid,
                        area=self.owner.area,
                    )
                )
                for date, time in self.owner.datetimes()
            ],
        )

    @cached_property
    def fields_ml(self):
        param, level = self.owner.param_level_ml
        if not (param and level):
            return ekd.from_source("empty")

        LOG.info(f"Loading model fields from {self.WHERE}")
        return ekd.from_source(
            "multi",
            [
                self.ml_load_source(
                    **self._patch(
                        date=date,
                        time=time,
                        param=param,
                        level=level,
                        grid=self.owner.grid,
                        area=self.owner.area,
                    )
                )
                for date, time in self.owner.datetimes()
            ],
        )

    @cached_property
    def all_fields(self):
        return self.fields_sfc + self.fields_pl + self.fields_ml



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property

import earthkit.data as ekd
import entrypoints

LOG = logging.getLogger(__name__)


class FileInput:
    def __init__(self, owner, file, **kwargs):
        self.file = file
        self.owner = owner

    @cached_property
    def fields_sfc(self):
        return self.all_fields.sel(levtype="sfc")

    @cached_property
    def fields_pl(self):
        return self.all_fields.sel(levtype="pl")

    @cached_property
    def fields_ml(self):
        return self.all_fields.sel(levtype="ml")

    @cached_property
    def all_fields(self):
        return ekd.from_source("file", self.file)


def get_input(name, *args, **kwargs):
    return available_inputs()[name].load()(*args, **kwargs)


def available_inputs():
    result = {}
    for e in entrypoints.get_group_all("ai_models_gfs.input"):
        result[e.name] = e
    return result
